use super::{
    block_cache_sync_all, get_block_cache, set_start_sec, BlockDevice, FSInfo, BLOCK_SZ, FAT,
};
// use crate::{dir_entry::*, layout::*, Inode};
use crate::{dir_entry::*, layout::*, println, Inode};
use alloc::string::{String, ToString};
use alloc::sync::Arc;
use alloc::vec::Vec;
use lazy_static::*;
use spin::{Mutex, RwLock};

/// 由于 fsinfo 结构只保存 fsinfo 扇区地址，所有的操作都通过查找相应的块完成，
/// 因此可以不对 fsinfo 添加锁。
///
/// 而 fat 结构在操作时还经过一些其它处理计算，当 fat 没有加锁，
/// 只对块缓冲加锁并且其操作完成时，可能转到其它线程执行，
/// 此时会破坏已有的结构（如当前请求的块缓冲被使用而导致继续使用该块的结果进行计算
/// 而产生错误），因此对 fat 进行了加锁。
pub struct FAT32Manager {
    block_device: Arc<dyn BlockDevice>,
    fsinfo: Arc<FSInfo>,
    fat: Arc<RwLock<FAT>>,
    pub sectors_per_cluster: u32,
    pub bytes_per_sector: u32,
    pub bytes_per_cluster: u32,
    root_sec: u32, // 数据区的起始扇区
}

type DataBlock = [u8; BLOCK_SZ];

impl FAT32Manager {
    /* 某个簇的第一个扇区 */
    pub fn first_sector_of_cluster(&self, cluster: u32) -> usize {
        (cluster as usize - 2) * self.sectors_per_cluster as usize + self.root_sec as usize
    }

    /* 打开现有的FAT32  */
    pub fn open(block_device: Arc<dyn BlockDevice>) -> Arc<RwLock<Self>> {
        // 读入分区偏移
        println!("[fs]: Load FAT32");
        let start_sector: u32 =
            get_block_cache(0, Arc::clone(&block_device))
                .lock()
                .read(0x1c6, |sector: &[u8; 4]| {
                    let mut start_sec: u32 = 0;
                    for i in 0..4 {
                        let tmp = sector[i] as u32;
                        start_sec = start_sec + (tmp << (8 * i));
                    }
                    start_sec
                });
        println!("[fs]: start sector = {}", start_sector);
        set_start_sec(start_sector as usize);

        // 读入 BPB
        let dbr = get_block_cache(0, Arc::clone(&block_device))
            .lock()
            .read(0, |dbr: &DataBlock| *dbr);
        let bpb = BPB::new(&dbr);
        println!("[fs]: BPB = {:?}", bpb);

        // 读入 FSInfo
        let fsinfo = FSInfo::new(bpb.fsinfo_sector as u32);
        // 校验签名
        assert!(
            fsinfo.check_signature(Arc::clone(&block_device)),
            "Error loading fat32! Illegal signature"
        );
        println!(
            "[fs]: first free cluster = {}",
            fsinfo.read_first_free_cluster(block_device.clone())
        );

        let sectors_per_cluster = bpb.sectors_per_cluster as u32;
        let bytes_per_sector = bpb.bytes_per_sector as u32;
        let bytes_per_cluster = sectors_per_cluster * bytes_per_sector;

        println!("[fs]: bytes per sec = {}", bytes_per_sector);
        println!("[fs]: bytes per cluster = {}", bytes_per_cluster);

        let fat_n_sec = bpb.sectors_per_fat;
        let fat1_sector = bpb.reserved_sectors as u32;
        let fat_n_entry = fat_n_sec * bytes_per_sector / 4;
        let fat = FAT::new(fat1_sector, fat_n_sec, fat_n_entry);

        println!(
            "[fs]: chain of root dir: {:?}",
            fat.get_all_clusters(2, Arc::clone(&block_device))
        );

        // 保留扇区数+所有FAT表的扇区数
        let root_sec = bpb.fats as u32 * fat_n_sec + bpb.reserved_sectors as u32;

        Arc::new(RwLock::new(Self {
            block_device,
            fsinfo: Arc::new(fsinfo),
            sectors_per_cluster,
            bytes_per_sector,
            bytes_per_cluster,
            fat: Arc::new(RwLock::new(fat)),
            root_sec,
        }))
    }

    /// 获取根节点
    pub fn root_inode(fs: &Arc<RwLock<Self>>) -> Inode {
        let (primary, ext) = fs.read().short_name_format("/");
        Inode::new(
            0,
            0,
            ShortDirEntry::new(primary, ext, Attribute::DIRECTORY.bits(), 2),
            Arc::clone(fs),
            Arc::clone(&fs.read().fat),
            Arc::clone(&fs.read().block_device),
        )
    }

    pub fn clear_cluster(&self, cluster_id: u32) {
        let start_sec = self.first_sector_of_cluster(cluster_id);
        for i in 0..self.sectors_per_cluster {
            get_block_cache(start_sec + i as usize, self.block_device.clone())
                .lock()
                .modify(0, |blk: &mut DataBlock| *blk = [0; BLOCK_SZ]);
        }
    }

    /// 分配簇，会填写FAT，成功返回第一个簇号，失败返回 0
    pub fn alloc_cluster(&self, num: u32) -> u32 {
        let free_clusters = self
            .fsinfo
            .read_free_clusters(Arc::clone(&self.block_device));
        if num > free_clusters {
            return 0;
        }

        // 获取FAT写锁
        let fat_writer = self.fat.write();
        let prev_cluster = self
            .fsinfo
            .read_first_free_cluster(self.block_device.clone());

        let first_cluster: u32 =
            fat_writer.next_free_cluster(prev_cluster, self.block_device.clone());
        let mut current_cluster = first_cluster;

        // 搜索可用簇，同时写表项
        for _ in 1..num {
            self.clear_cluster(current_cluster);
            let next_cluster =
                fat_writer.next_free_cluster(current_cluster, self.block_device.clone());
            assert_ne!(next_cluster, 0);
            fat_writer.set_next_cluster(current_cluster, next_cluster, self.block_device.clone());
            current_cluster = next_cluster;
        }

        self.clear_cluster(current_cluster);
        // 填写最后一个表项
        fat_writer.set_end(current_cluster, self.block_device.clone());
        // 修改FSINFO
        self.fsinfo
            .write_free_clusters(free_clusters - num, self.block_device.clone());
        // 写入分配的最后一个簇
        self.fsinfo
            .write_first_free_cluster(current_cluster, self.block_device.clone());
        self.cache_write_back();

        first_cluster
    }

    pub fn dealloc_cluster(&self, clusters: Vec<u32>) {
        let fat_writer = self.fat.write();
        let free_clusters = self
            .fsinfo
            .read_free_clusters(Arc::clone(&self.block_device));
        let num = clusters.len();
        for i in 0..num {
            // 将FAT对应表项清零
            fat_writer.set_next_cluster(clusters[i], FREE_CLUSTER, self.block_device.clone())
        }
        // 修改FSINFO
        if num > 0 {
            self.fsinfo
                .write_free_clusters(free_clusters + num as u32, self.block_device.clone());
            // 如果释放的簇号小于开始空闲簇字段，更新该字段
            if clusters[0] > 2
                && clusters[0]
                    < self
                        .fsinfo
                        .read_first_free_cluster(self.block_device.clone())
            {
                self.fsinfo
                    .write_first_free_cluster(clusters[0] - 1, self.block_device.clone());
            }
        }
    }

    /// 字节转化为所需的簇数
    fn size_to_clusters(&self, size: usize) -> usize {
        (size + self.bytes_per_cluster as usize - 1) / self.bytes_per_cluster as usize
    }

    /// 计算扩大至 new_size(B) 需要多少个簇
    pub fn cluster_num_needed(&self, old_size: usize, new_size: usize) -> usize {
        if old_size >= new_size {
            0
        } else {
            self.size_to_clusters(new_size) - self.size_to_clusters(old_size)
        }
    }

    /* 计算当前偏移量在第几个簇 */
    pub fn cluster_of_offset(&self, offset: usize) -> u32 {
        offset as u32 / self.bytes_per_cluster
    }

    /// 获取偏移量所在的簇、扇区和偏移
    /// 
    /// first_cluster: 首簇，起始位置
    pub fn get_pos(&self, first_cluster: u32, offset: usize) -> (u32, usize, usize) {
        let fat = self.fat.read();
        let cluster_index = self.cluster_of_offset(offset);
        let current_cluster =
            fat.get_cluster_at(first_cluster, cluster_index, Arc::clone(&self.block_device));
        let current_sector = self.first_sector_of_cluster(current_cluster)
            + (offset - cluster_index as usize * self.bytes_per_cluster as usize)
                / self.bytes_per_sector as usize;
        (
            current_cluster,
            current_sector,
            offset % self.bytes_per_sector as usize,
        )
    }

    pub fn cache_write_back(&self) {
        block_cache_sync_all();
    }
}

/// 该 impl 块实现与长短文件名有关的操作
impl FAT32Manager {
    /* 拆分文件名和后缀 */
    pub fn split_name_ext<'a>(&self, name: &'a str) -> (&'a str, &'a str) {
        let mut primary_and_ext: Vec<&str> = name.split(".").collect();
        let primary = primary_and_ext[0];
        if primary_and_ext.len() == 1 {
            primary_and_ext.push("");
        }
        let ext = primary_and_ext[primary_and_ext.len() - 1];
        (primary, ext)
    }

    /// 判断一个文件名是否符合短文件名
    pub fn is_short_name(&self, name: &str) -> bool {
        let (primary, ext) = self.split_name_ext(name);
        if primary.len() <= SHORT_PRIMARY_LEN && ext.len() <= SHORT_EXT_LEN {
            true
        } else {
            false
        }
    }

    /// 将短文件名格式化为目录项存储的内容
    pub fn short_name_format(&self, name: &str) -> ([u8; 8], [u8; 3]) {
        let name = &name.to_ascii_uppercase()[..];
        let (mut primary, mut ext) = self.split_name_ext(name);
        if name == "." || name == ".." {
            primary = name;
            ext = ""
        }
        let primary_bytes = primary.as_bytes();
        let ext_bytes = ext.as_bytes();
        let mut f_name = [0x20; 8];
        let mut f_ext = [0x20; 3];
        f_name[0..primary_bytes.len()].copy_from_slice(primary_bytes);
        f_ext[0..ext_bytes.len()].copy_from_slice(ext_bytes);
        (f_name, f_ext)
    }

    /// 将长文件名拆分格式化，末尾补0，其余填补0xff
    pub fn long_name_format(&self, name: &str) -> Vec<[u8; 13]> {
        let len = name.len() as u32;
        let name = name.as_bytes();
        let num = (len + LONG_NAME_LEN - 1) / LONG_NAME_LEN;
        let mut res: Vec<[u8; 13]> = Vec::new();
        for i in 0..num {
            let mut tmp = [0xff; 13];
            let interval = LONG_NAME_LEN.min(len - i * LONG_NAME_LEN);
            for j in 0..interval {
                tmp[j as usize] = name[(i * LONG_NAME_LEN + j) as usize];
            }
            // 长文件名末尾需要补0
            if interval != LONG_NAME_LEN {
                tmp[interval as usize] = 0;
            }
            res.push(tmp);
        }
        res
    }

    /// 由长文件名生成短文件名
    /// 只支持 ASCII
    pub fn generate_short_name(&self, long_name: &str) -> String {
        let mut res = String::new();
        let numeric_tail = numeric_alloc();
        let long_name = long_name.trim_matches(|c| c == '\n' || c == ' ' || c == '.');
        let split = long_name.split('.').collect::<Vec<_>>();

        // 构造 primary name
        for c in split[0].as_bytes() {
            if res.len() >= 8 - numeric_tail.len() {
                break;
            }
            if *c == ' ' as u8 {
                continue;
            }
            res.push((*c as char).to_ascii_uppercase());
        }
        res.push_str(&numeric_tail);

        // 构造 ext
        if split.len() > 1 {
            res.push('.');
            let mut i = 0;
            for c in split[split.len() - 1].as_bytes() {
                if i >= 3 {
                    break;
                }
                if *c == ' ' as u8 {
                    continue;
                }
                res.push((*c as char).to_ascii_uppercase());
                i += 1;
            }
        }

        res
    }
}

struct NumericAllocator {
    num: usize,
}

impl NumericAllocator {
    pub fn new() -> Self {
        NumericAllocator { num: 0 }
    }
    pub fn alloc(&mut self) -> String {
        self.num += 1;
        String::from(String::from("~") + &self.num.to_string())
    }
}

lazy_static! {
    static ref NUMERIC_ALLOCATOR: Mutex<NumericAllocator> = Mutex::new(NumericAllocator::new());
}

pub fn numeric_alloc() -> String {
    NUMERIC_ALLOCATOR.lock().alloc()
}
