use alloc::vec::Vec;
use spin::{Mutex, MutexGuard};

use super::FileClass;

pub struct Files {
    inner: Mutex<FilesInner>,
}

pub struct FilesInner {
    table: Vec<Option<FileClass>>,
}

impl Files {
    pub fn inner_exclusive_access(&self) -> MutexGuard<'_, FilesInner> {
        self.inner.lock()
    }

    pub fn empty() -> Self {
        Self {
            inner: Mutex::new(FilesInner { table: Vec::new() }),
        }
    }

    pub fn new(table: Vec<Option<FileClass>>) -> Self {
        Self {
            inner: Mutex::new(FilesInner { table }),
        }
    }

    pub fn len(&self) -> usize {
        self.inner_exclusive_access().table.len()
    }

    pub fn get(&self, fd: usize) -> Option<FileClass> {
        let inner = self.inner_exclusive_access();
        inner
            .table
            .get(fd)
            .map(|f| f.as_ref())
            .flatten()
            .map(|f| f.clone())
    }

    pub fn alloc_fd(&self) -> usize {
        let mut inner = self.inner_exclusive_access();
        if let Some(fd) = (0..inner.table.len()).find(|fd| inner.table[*fd].is_none()) {
            fd
        } else {
            inner.table.push(None);
            inner.table.len() - 1
        }
    }

    /// 将文件描述符与文件绑定
    pub fn set_fd(&self, fd: usize, file: FileClass) {
        let mut inner = self.inner_exclusive_access();
        while fd > inner.table.len() - 1 {
            inner.table.push(None);
        }
        inner.table[fd] = Some(file);
    }

    pub fn remove(&self, fd: usize) -> Option<FileClass> {
        let mut inner = self.inner_exclusive_access();
        inner.table.push(None);
        inner.table.swap_remove(fd)
    }

    pub fn clone(&self) -> Self {
        let inner = self.inner_exclusive_access();
        Self {
            inner: Mutex::new(FilesInner {
                table: Vec::clone(&inner.table),
            }),
        }
    }

    pub fn clear(&self) {
        let mut inner = self.inner_exclusive_access();
        inner.table.clear();
    }
}
