mod files;
mod info;
mod inode;
mod pipe;
mod stdio;

use crate::drivers::BLOCK_DEVICE;
use crate::mm::UserBuffer;
use alloc::sync::Arc;
use alloc::vec::Vec;
use fat32::{FAT32Manager, Inode, END_CLUSTER};
use lazy_static::*;
use spin::{Mutex, RwLock};

pub trait File: Send + Sync {
    fn readable(&self) -> bool;
    fn writable(&self) -> bool;
    fn read(&self, buf: UserBuffer) -> usize;
    fn write(&self, buf: UserBuffer) -> usize;
}

#[derive(Clone)]
pub enum FileClass {
    OSInode(Arc<OSInode>),
    PIPE(Arc<Pipe>),
    Other(Arc<dyn File + Send + Sync>),
}

pub use files::Files;
pub use info::{Dirent, IOVec, Kstat, Winsize};
pub use inode::*;
pub use pipe::{make_pipe, Pipe};
pub use stdio::{Stdin, Stdout};

lazy_static! {
    /// inode 表，整个系统只有一份，存储文件的 inode 结点，且每个结点只有一份
    ///
    /// 如果该表在 vfs.rs 中进行管理，则什么时候进行创建或删除无法确定，
    /// 因此可以在需要打开文件的时候进行处理，这时能够保证所获得的文件都是需要的，且应该加进表中
    static ref INODE_TABLE: Mutex<Vec<Option<Arc<RwLock<Inode>>>>> = Mutex::new(Vec::new());

    /// 打开文件表，整个系统只有一份，存储打开的文件的 OSInode 结点
    ///
    /// 可以打开相同的文件，这时会在打开文件表存储两个 OSInode 结点，
    /// 但这两个结点指向的 Inode 结点相同，只是文件偏移量不同。</br>
    /// 根结点存在 0 号位置
    static ref FILE_TABLE: Mutex<Vec<Option<Arc<OSInode>>>> = Mutex::new(Vec::new());

    pub static ref ROOT_INODE: Arc<OSInode> = {
        let fs = FAT32Manager::open(BLOCK_DEVICE.clone());
        println!("[fs]: open fat32 success!");
        let root_inode = FAT32Manager::root_inode(&fs);
        println!("[fs]: get root indoe success!");
        println!("root first cluster = {}", root_inode.first_cluster());
        insert_file_table(true, true, root_inode)
    };
}

/// 向 inode 表插入结点
pub fn insert_inode_table(inode: Inode) -> Arc<RwLock<Inode>> {
    // println!("in insert_inode_table");
    let mut inode_table = INODE_TABLE.lock();
    let first_cluster = inode.first_cluster();
    for i in 0..inode_table.len() {
        if inode_table[i].is_some() {
            let tmp = Arc::clone(&inode_table[i].as_ref().unwrap());
            let read_tmp = tmp.read();
            // println!(
            //     "index = {}, source first cluste = {}, dst first cluster = {}",
            //     i,
            //     read_tmp.first_cluster(),
            //     inode.first_cluster()
            // );
            if read_tmp.first_cluster() == first_cluster
                && (first_cluster != END_CLUSTER || read_tmp.path() == inode.path())
            {
                // println!("get from inode table: index = {}", i);
                return Arc::clone(&tmp);
            }
            drop(read_tmp);
        }
    }
    // let name = inode.name();
    let res = Arc::new(RwLock::new(inode));
    if let Some(num) = (0..inode_table.len()).find(|num| inode_table[*num].is_none()) {
        // println!("insert inode table: index = {}, name = {}", num, name);
        inode_table[num] = Some(Arc::clone(&res));
    } else {
        // println!(
        //     "insert inode table: index = {}, name = {}",
        //     inode_table.len(),
        //     name
        // );
        inode_table.push(Some(Arc::clone(&res)));
    }
    res
}

/// 向打开文件表插入结点
pub fn insert_file_table(readable: bool, writable: bool, inode: Inode) -> Arc<OSInode> {
    // println!("in insert_file_table");
    let mut file_table = FILE_TABLE.lock();
    let res = Arc::new(OSInode::new(readable, writable, insert_inode_table(inode)));
    if let Some(num) = (0..file_table.len()).find(|num| file_table[*num].is_none()) {
        file_table[num] = Some(Arc::clone(&res));
    } else {
        file_table.push(Some(Arc::clone(&res)));
    }
    res
}

/// 移除 inode 表不被使用的结点
pub fn remove_from_inode_table() {
    let mut inode_table = INODE_TABLE.lock();
    loop {
        let mut num = 0;
        while num < inode_table.len() {
            if inode_table[num].is_some()
                && Arc::strong_count(&inode_table[num].as_ref().unwrap()) == 1
            {
                // let tmp = Arc::clone(&inode_table[num].as_ref().unwrap());
                // println!(
                //     "in remove_from_inode_table: delete index = {}",
                //     tmp.read().name()
                // );
                // drop(tmp);
                inode_table.remove(num).unwrap();
                break;
            }
            num += 1;
        }
        if num == inode_table.len() {
            return;
        }
    }
}

/// 移除打开文件表不被使用的结点
pub fn remove_from_file_table() {
    let mut file_table = FILE_TABLE.lock();
    loop {
        let mut num = 0;
        while num < file_table.len() {
            if file_table[num].is_some()
                && Arc::strong_count(&file_table[num].as_ref().unwrap()) == 1
            {
                // let tmp = Arc::clone(&file_table[num].as_ref().unwrap());
                // println!(
                //     "in remove_from_file_table: delete name = {}",
                //     tmp.inner_exclusive_access().inode.read().name()
                // );
                // drop(tmp);
                file_table.remove(num).unwrap();
                break;
            }
            num += 1;
        }
        if num == file_table.len() {
            remove_from_inode_table();
            return;
        }
    }
}
