use alloc::vec::Vec;

use crate::{
    fs::{File, FileClass},
    mm::{MapPermission, PhysAddr, UserBuffer, VirtAddr},
    task::current_task,
};

pub fn sys_brk(addr: usize) -> isize {
    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    let mm = &task_inner.memory_set;
    let brk = mm.inner_exclusive_access().brk;
    if addr != 0 {
        let permission = MapPermission::U | MapPermission::R | MapPermission::W;
        let new_brk = VirtAddr::from(addr).ceil().into();
        mm.insert_framed_area(brk, new_brk, permission);
        mm.inner_exclusive_access().brk = new_brk;
        usize::from(new_brk) as isize
    } else {
        usize::from(brk) as isize
    }
}

bitflags! {
    pub struct ProtFlags: u32 {
        const PROT_NONE = 0;
        const PROT_READ = 0x1;
        const PROT_WRITE = 0x2;
        const PROT_EXEC = 0x4;
    }
}

bitflags! {
    pub struct MapFlags: u32 {
        const MAP_SHARED = 0x1;
        const MAP_PRIVATE = 0x2;
        // MAP_FIXED 表示直接使用传递的参数做为 mmap 空间的起始地址
        const MAP_FIXED = 0x10;
        const MAP_ANONYMOUS = 0x20;
    }
}

/// 只能读取文件，不能写回文件
pub fn sys_mmap(start: usize, len: usize, prot: u32, flags: u32, fd: usize, off: usize) -> isize {
    let prot = ProtFlags::from_bits(prot).unwrap();
    let flags = MapFlags::from_bits(flags).unwrap();
    // println!(
    //     "in sys_mmap: start = {:#x}, len = {:#x}, prot = {:?}, flags = {:?}, fd = {}, off = {:#x}",
    //     start, len, prot, flags, fd as isize, off
    // );
    let mut permission = MapPermission::U;
    if prot.contains(ProtFlags::PROT_READ) {
        permission |= MapPermission::R;
    }
    if prot.contains(ProtFlags::PROT_WRITE) {
        permission |= MapPermission::W;
    }
    if prot.contains(ProtFlags::PROT_EXEC) {
        permission |= MapPermission::X;
    }

    let task = current_task().unwrap();
    let task_inner = task.inner_exclusive_access();
    // 分配 mmap 空间
    // 可能有的重用之前已经分配的空间，就造成 mmap 空间的分割
    let mut res = start;
    if flags.contains(MapFlags::MAP_FIXED) {
        task_inner
            .memory_set
            .insert_framed_area(start.into(), (start + len).into(), permission);
    } else {
        res = task_inner
            .memory_set
            .insert_framed_area_randomly(len, None, permission)
            .0
            .into();
    }

    // 匿名映射不关联任何文件
    if flags.contains(MapFlags::MAP_ANONYMOUS) {
        res as isize
    } else {
        if let Some(inode) = task_inner.fd_table.get(fd) {
            match inode {
                FileClass::OSInode(f) => {
                    // println!(
                    //     "inode name = {}",
                    //     f.inner_exclusive_access().inode.read().name()
                    // );

                    let ppn: PhysAddr = task_inner
                        .memory_set
                        .translate(VirtAddr::from(res).into())
                        .unwrap()
                        .ppn()
                        .into();

                    let mut vec: Vec<&'static mut [u8]> = Vec::new();
                    vec.push(unsafe {
                        core::slice::from_raw_parts_mut(usize::from(ppn) as *mut u8, len)
                    });

                    f.seek(off);
                    f.read(UserBuffer::new(vec));
                    return res as isize;
                }
                _ => {
                    return -1;
                }
            }
        }
        -1
    }
}

pub fn sys_mprotect(addr: usize, len: usize, prot: u32) -> isize {
    let prot = ProtFlags::from_bits(prot).unwrap();
    let task = current_task().unwrap();
    let mut permission = MapPermission::U;
    if prot.contains(ProtFlags::PROT_READ) {
        permission |= MapPermission::R;
    }
    if prot.contains(ProtFlags::PROT_WRITE) {
        permission |= MapPermission::W;
    }
    if prot.contains(ProtFlags::PROT_EXEC) {
        permission |= MapPermission::X;
    }
    task.inner_exclusive_access().memory_set.insert_framed_area(
        addr.into(),
        (addr + len).into(),
        permission,
    );
    0
}

pub fn sys_munmap(start: usize, _len: usize) -> isize {
    let task = current_task().unwrap();
    task.inner_exclusive_access()
        .memory_set
        .remove_area_with_start_vpn(start.into());
    0
}
