use super::errno::*;
use crate::fs::{open_file_at, OpenFlags, AT_FDCWD};
use crate::mm::{
    translated_byte_buffer, translated_ref, translated_refmut, translated_str, UserBuffer, VirtAddr,
};
use crate::task::{
    current_task, current_user_token, exit_current_and_run_next, exit_current_group,
    move_child_process_to_init, remove_from_thread_group, suspend_current_and_run_next,
    TaskControlBlock, PIDTYPE_PPID, PID_HASH,
};
use crate::timer::*;
use crate::utsname::*;
use alloc::string::String;
use alloc::sync::Arc;
use alloc::vec::Vec;
use core::mem::size_of;

pub fn sys_times(buf: *mut Tms) -> isize {
    let token = current_user_token();
    let res = get_time();
    (*translated_refmut(token, buf)).tms_utime = res as u64;
    (*translated_refmut(token, buf)).tms_stime = res as u64;
    (*translated_refmut(token, buf)).tms_cutime = res as u64;
    (*translated_refmut(token, buf)).tms_cstime = res as u64;
    0
}

pub fn sys_get_time_of_day(time: *mut Timeval) -> isize {
    let token = current_user_token();
    (*translated_refmut(token, time)).tv_sec = get_time_s() as u64;
    (*translated_refmut(token, time)).tv_usec = get_time_us() as u64;
    0
}

pub fn sys_clock_gettime(_clk_id: usize, tp: *mut Timespec) -> isize {
    let token = current_user_token();
    (*translated_refmut(token, tp)).tv_sec = get_time_s();
    (*translated_refmut(token, tp)).tv_nsec = get_time_ns();
    0
}

pub fn sys_nanosleep(req: *mut Timespec, _rem: *mut Timespec) -> isize {
    let token = current_user_token();
    let end_sec = get_time_s() + (*translated_refmut(token, req)).tv_sec as usize;
    let end_usec = get_time_ns() + (*translated_refmut(token, req)).tv_nsec as usize;
    loop {
        let cur_sec = get_time_s();
        let cur_usec = get_time_ns();
        if cur_sec >= end_sec && cur_usec >= end_usec {
            return 0;
        } else {
            suspend_current_and_run_next();
        }
    }
}

pub fn sys_uname(buf: *mut Utsname) -> isize {
    let token = current_user_token();
    let mut buffer = UserBuffer::new(translated_byte_buffer(
        token,
        buf as usize as *const u8,
        size_of::<Utsname>(),
    ));
    let uname = Utsname::new();
    buffer.read(uname.as_bytes());
    0
}

pub fn sys_gettid() -> isize {
    current_task().unwrap().get_tid() as isize
}

pub fn sys_getpid() -> isize {
    current_task().unwrap().get_pid() as isize
}

pub fn sys_getppid() -> isize {
    current_task().unwrap().get_ppid() as isize
}

pub fn sys_getuid() -> isize {
    0
}

pub fn sys_exit(exit_code: i32) -> ! {
    exit_current_and_run_next(exit_code);
    panic!("Unreachable in sys_exit!");
}

pub fn sys_exit_group(exit_code: i32) -> ! {
    exit_current_group(exit_code);
    panic!("Unreachable in sys_exit!");
}

pub fn sys_sched_yield() -> isize {
    suspend_current_and_run_next();
    0
}

bitflags! {
    pub struct CloneFlags: u32 {
        const CSIGNAL = 0x000000ff;
        // const CLONE_NEWTIME= 0x00000080;
        const CLONE_VM = 0x00000100;
        const CLONE_FS = 0x00000200;
        const CLONE_FILES = 0x00000400;
        const CLONE_SIGHAND = 0x00000800;
        // const CLONE_PIDFD = 0x00001000;
        // const CLONE_PTRACE = 0x00002000;
        // const CLONE_VFORK =0x00004000;
        // const CLONE_PARENT = 0x00008000;
        const CLONE_THREAD = 0x00010000;
        // const CLONE_NEWNS = 0x00020000;
        const CLONE_SYSVSEM = 0x00040000;
        const CLONE_SETTLS = 0x00080000;
        const CLONE_PARENT_SETTID = 0x00100000;
        const CLONE_CHILD_CLEARTID = 0x00200000;
        const CLONE_DETACHED = 0x00400000;
        // const CLONE_UNTRACED = 0x00800000;
        const CLONE_CHILD_SETTID = 0x01000000;
        // const CLONE_NEWCGROUP = 0x02000000;
        // const CLONE_NEWUTS = 0x04000000;
        // const CLONE_NEWIPC = 0x08000000;
        // const CLONE_NEWUSER = 0x10000000;
        // const CLONE_NEWPID = 0x20000000;
        // const CLONE_NEWNET = 0x40000000;
        // const CLONE_IO = 0x80000000;
    }
}

pub fn sys_clone(
    flags: u32,
    stack: usize,
    parent_tid: *mut u32,
    tls: usize,
    child_tid: *mut u32,
) -> isize {
    let flags = CloneFlags::from_bits(flags).unwrap();
    // println!("clone: flags = {:#?}", flags);
    let current_process = current_task().unwrap();
    let stack = if stack == 0 {
        None
    } else {
        Some(VirtAddr::from(stack))
    };
    let new_process =
        TaskControlBlock::clone(&current_process, flags, stack, parent_tid, tls, child_tid);
    let new_pid = new_process.get_pid();
    // modify trap context of new_process, because it returns immediately after switching
    let new_process_inner = new_process.inner_exclusive_access();
    let trap_cx = new_process_inner.get_trap_cx();
    // we do not have to move to next instruction since we have done it before
    // for child process, fork returns 0
    trap_cx.x[10] = 0;
    new_pid as isize
}

pub fn sys_execve(path: *const u8, mut args: *const usize, mut envp: *const usize) -> isize {
    let token = current_user_token();
    let path = translated_str(token, path);
    let mut args_vec: Vec<String> = Vec::new();
    loop {
        let arg_str_ptr = *translated_ref(token, args);
        if arg_str_ptr == 0 {
            break;
        }
        args_vec.push(translated_str(token, arg_str_ptr as *const u8));
        unsafe {
            args = args.add(1);
        }
    }
    let mut envp_vec: Vec<String> = Vec::new();
    loop {
        let env_str_ptr = *translated_ref(token, envp);
        if env_str_ptr == 0 {
            break;
        }
        envp_vec.push(translated_str(token, env_str_ptr as *const u8));
        unsafe {
            envp = envp.add(1);
        }
    }
    // println!("in execve: app name = {}, argv = {:?}, envp = {:?}", path.as_str(), args_vec, envp_vec);
    if let Some(app_inode) = open_file_at(AT_FDCWD, path.as_str(), OpenFlags::RDONLY) {
        let all_data = app_inode.read_all();
        let process = current_task().unwrap();
        process.exec(all_data.as_slice(), args_vec, envp_vec);
        // println!("exec successfully");
        0
    } else {
        -ENOENT
    }
}

bitflags! {
    pub struct WaitFlags: u32 {
        const WNOHANG = 0x00000001;
        // const WUNTRACED = 0x00000002;
        // const WCONTINUED = 0x00000008;
    }
}

pub fn sys_wait4(pid: isize, wstatus: *mut i32, flags: u32, _rusage: usize) -> isize {
    let flags = WaitFlags::from_bits(flags).unwrap();
    loop {
        let mut hash = PID_HASH[PIDTYPE_PPID].lock();
        let task = current_task().unwrap();
        let task_inner = task.inner_exclusive_access();

        let children = hash.get_mut(&task.get_pid()).unwrap();
        let mut wait_child_id = None;
        let mut exit_child = None;
        for (id, child) in children.iter().enumerate() {
            // 检查子线程组是否退出
            let child_inner = child.inner_exclusive_access();
            if child_inner
                .signal_info
                .inner_exclusive_access()
                .group_exit_code
                .is_some()
            {
                exit_child = Some(Arc::clone(child));
            }
            drop(child_inner);

            // 检查子线程组是否满足要求
            match pid {
                _ if pid > 0 => {
                    if child.get_pid() == pid as usize {
                        wait_child_id = Some(id);
                        break;
                    }
                }
                _ => {
                    wait_child_id = Some(id);
                    if exit_child.is_some() {
                        break;
                    }
                }
            }
        }

        if let Some(id) = wait_child_id {
            if let Some(child) = exit_child {
                let pid = child.get_pid();
                let child_inner = child.inner_exclusive_access();
                let exic_code = child_inner
                    .signal_info
                    .inner_exclusive_access()
                    .group_exit_code
                    .unwrap();
                // C 库中 #define WEXITSTATUS(s) ((s & 0xFF00) >> 8)
                // exit_code 需要左移 8 位
                *translated_refmut(task_inner.get_user_token(), wstatus) = exic_code << 8;
                drop(child_inner);

                children.remove(id);
                remove_from_thread_group(pid);
                // 调用 wait4() 的可能是 init 进程，因此需要释放进程锁
                drop(task_inner);
                drop(task);
                drop(hash);
                move_child_process_to_init(pid);

                // 此时 TCB 引用应该只剩一个，可以验证一下
                assert_eq!(Arc::strong_count(&child), 1);
                return pid as isize;
            } else if flags.contains(WaitFlags::WNOHANG) {
                return 0;
            } else {
                drop(task_inner);
                drop(task);
                drop(hash);
                suspend_current_and_run_next();
            }
        } else {
            return -ECHILD;
        }
    }
}

pub fn sys_set_tid_address(ptr: usize) -> isize {
    let task = current_task().unwrap();
    let mut task_inner = task.inner_exclusive_access();
    task_inner.clear_child_tid = Some(ptr.into());
    task.get_tid() as isize
}
